/*
  64 memory handling
*/

#include "memhand.h"
#include <malloc.h>
#include <string.h>
#include <stdio.h>

char *g_n64rom;
char boot[0x2000];
unsigned int size;
char RDram_Registers[0x2F];
char *RDram;
char *SP_Registers;             // 0400000h     // Boot?
char DP_Cmd_Registers[0x0F];    // 0410000h
char DP_Span_Registers[0x0F];   // 0420000h
char MI_Registers[0x0F];        // 0430000h
char VI_Registers[0x37];        // 0440000h
char AI_Registers[0x17];        // 0450000h
char PI_Registers[0x34];        // 0460000h
char RI_Registers[0x1F];        // 0470000h
char SI_Registers[0x1B];        // 0480000h
char PIF_RxM[0x7FF];
PI_typ PI;

// Init Memory
//////////////////////////////////////////////////////////////////////////
void init_mem(void)
{
    DWORD *temp;

    RDram = (char*)malloc(0x400000);
#ifdef GP2X
    gp2x_memset(&RDram[0], 0, 0x400000);
    gp2x_memset(&RDram_Registers[0],0, 0x2f);
    gp2x_memset(&DP_Cmd_Registers[0],0, 0x0f);
    gp2x_memset(&DP_Span_Registers[0],0, 0x0f);
    gp2x_memset(&MI_Registers[0], 0, 0x0f);
    gp2x_memset(&VI_Registers[0], 0, 0x37);
    gp2x_memset(&AI_Registers[0], 0, 0x17);
    gp2x_memset(&PI_Registers[0], 0, 0x33);
    gp2x_memset(&RI_Registers[0], 0, 0x1f);
    gp2x_memset(&SI_Registers[0], 0, 0x1b);
#else
    memset(&RDram[0], 0, 0x400000);
    memset(&RDram_Registers[0],0, 0x2f);
    memset(&DP_Cmd_Registers[0],0, 0x0f);
    memset(&DP_Span_Registers[0],0, 0x0f);
    memset(&MI_Registers[0], 0, 0x0f);
    memset(&VI_Registers[0], 0, 0x37);
    memset(&AI_Registers[0], 0, 0x17);
    memset(&PI_Registers[0], 0, 0x33);
    memset(&RI_Registers[0], 0, 0x1f);
    memset(&SI_Registers[0], 0, 0x1b);
#endif

    temp = (DWORD*)&MI_Registers[4];
    *temp = 0x01010101;
}

// Mem Copy
//////////////////////////////////////////////////////////////////////////
inline void r_memcpy8( void* out, void* in )
{
	((char *)out)[0] = ((char*)in)[0];
}

inline void r_memcpy16( void* out, void* in )
{
	((char *)out)[0] = ((char*)in)[1];
	((char *)out)[1] = ((char*)in)[0];
}

inline void r_memcpy32( void* out, void* in )
{
	((char *)out)[0] = ((char*)in)[3];
	((char *)out)[1] = ((char*)in)[2];
	((char *)out)[2] = ((char*)in)[1];
	((char *)out)[3] = ((char*)in)[0];
}

inline void r_memcpy64( void* out, void* in )
{
	((char *)out)[0] = ((char*)in)[7];
	((char *)out)[1] = ((char*)in)[6];
	((char *)out)[2] = ((char*)in)[5];
	((char *)out)[3] = ((char*)in)[4];

	((char *)out)[4] = ((char*)in)[3];
	((char *)out)[5] = ((char*)in)[2];
	((char *)out)[6] = ((char*)in)[1];
	((char *)out)[7] = ((char*)in)[0];
}

inline void r_memcpy (void *out, void *in, size_t size)
{
  size_t x = --size;
  size_t y = 0;
  char *src = (char*)in;
  char *dest = (char*)out;

  for (; x != -1 ; --x)
  {
    dest[y++] = src[x];
  }
}

#define SWAP_WORD(N) (((N) << 24) |	\
                 (((N) << 8) & 0x00ff0000) | \
                 (((N) >> 8) & 0x0000ff00) | \
                 (((unsigned int)(N)) >> 24))

#define SWAP_DWORD(N) (((N & 0xff) << 24) | ((N & 0xff00) << 8) | \
		            ((N & 0xff0000) >> 8) | ((N & 0xff000000) >> 24))

// Do Pi Transfer
//////////////////////////////////////////////////////////////////////////
inline void Do_PI_Transfer(void)
{
  DWORD count;

#ifdef DEBUG
  printf("Doing PI transfer of %08X bytes from: %08X to %08X\n", PI.WR_LEN, PI.Cart, PI.DRAM);
#endif
  for (count = 0; count <= PI.WR_LEN; count++)
    putbyte(PI.DRAM+count,getbyte(PI.Cart+count));
}
int ka = 0;
// Read Memory
//////////////////////////////////////////////////////////////////////////
inline DWORD getdword (unsigned int location)
{
    DWORD ret = 0;

    #ifdef DEBUGMEM
    printf("** Reading from memory location %08X\n", location);
    #endif

    // Everything WRAPS!!!
    location &= 0x1FFFFFFF;

    if(location >= 0x1FC00000 && location <= 0x1FC007FF)
    {
        //return SWAP_WORD(((u32*)&PIF_RxM[location&0x7FF])[0]);
        r_memcpy32(&ret, &PIF_RxM[location & 0x7FF]);
        return(ret);
    }
    if(location >= 0x10000000)
    {
        //return SWAP_WORD(((u32*)&g_n64rom[location&0x0FFFFFFF])[0]);
        r_memcpy32(&ret, &g_n64rom[location & 0x0FFFFFFF]);
        return(ret);
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
            //return SWAP_WORD(((u32*)&SI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &SI_Registers[location & 0xff]);
              return(ret);
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
            //return SWAP_WORD(((u32*)&RI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &RI_Registers[location & 0xfF]);
              return(ret);
        }
        if(location >= 0x04600000 && location <= 0x04600033)
        {
            //return SWAP_WORD(((u32*)&PI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &PI_Registers[location & 0xff]);
              return(ret);
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
            //return SWAP_WORD(((u32*)&AI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &AI_Registers[location & 0xff]);
              return(ret);
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
            //return SWAP_WORD(((u32*)&VI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &VI_Registers[location & 0xff]);
              return(ret);
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
            //return SWAP_WORD(((u32*)&MI_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &MI_Registers[location & 0xfF]);
              return(ret);
        }

        if(location >= 0x04200000 && location <= 0x0420000F)
        {
            //return SWAP_WORD(((u32*)&DP_Span_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &DP_Span_Registers[location & 0xfF]);
              return(ret);
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
            //return SWAP_WORD(((u32*)&DP_Cmd_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &DP_Cmd_Registers[location & 0xfF]);
              return(ret);
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
            //return SWAP_WORD(((u32*)&boot[location&0x1FFF])[0]);
              r_memcpy32(&ret, &boot[location & 0x1FFF]);
              return(ret);
        }
        if(location >= 0x03f00000 && location <  0x03f0002f)
        {
            //return SWAP_WORD(((u32*)&RDram_Registers[location&0xFF])[0]);
              r_memcpy32(&ret, &RDram_Registers[location & 0xff]);
              return(ret);
        }
    }
    if(location <= 0x003fffff)
    {
        //return SWAP_WORD(((u32*)&RDram[location])[0]);
        r_memcpy32(&ret, &RDram[location]);
        return(ret);
    }
    return 0;
}

inline WORD getword(unsigned int location)
{
    WORD ret = 0;

    #ifdef DEBUGMEM
    printf("** Reading from memory location %08X\n", location);
    #endif

    location &= 0x1FFFFFFF; // Everything WRAPS!!!

    if(location >= 0x1FC00000 && location <= 0x1FC007FF)
    {
        //return SWAP_WORD(((u16*)&PIF_RxM[location&0x7FF])[0]);
        r_memcpy16(&ret, &PIF_RxM[location & 0x7FF]);
        return(ret);
    }
    if(location >= 0x10000000)
    {
        //return SWAP_WORD(((u16*)&g_n64rom[location&0x0FFFFFFF])[0]);
        r_memcpy16(&ret, &g_n64rom[location & 0x0FFFFFFF]);
        return(ret);
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
            //return SWAP_WORD(((u16*)&SI_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &SI_Registers[location & 0xff]);
            return(ret);
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
            //return SWAP_WORD(((u16*)&RI_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &RI_Registers[location & 0xfF]);
            return(ret);
        }
        if(location >= 0x04600000 && location <= 0x04600033)
        {
            //return SWAP_WORD(((u16*)&PI_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &PI_Registers[location & 0xff]);
            return(ret);
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
            //return SWAP_WORD(((u16*)&AI_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &AI_Registers[location & 0xff]);
            return(ret);
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
            //return SWAP_WORD(((u16*)&VI_Registers[location&0xFF])[0]);
            // ??
            //r_memcpy32(&ret, &VI_Registers[location & 0xff]);
            r_memcpy16(&ret, &VI_Registers[location & 0xff]);
            return(ret);
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
            //return SWAP_WORD(((u16*)&MI_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &MI_Registers[location & 0xfF]);
            return(ret);
        }
        if(location >= 0x04200000 && location <= 0x0420000F)
        {
            //return SWAP_WORD(((u16*)&DP_Span_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &DP_Span_Registers[location & 0xfF]);
            return(ret);
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
            //return SWAP_WORD(((u16*)&DP_Cmd_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &DP_Cmd_Registers[location & 0xfF]);
            return(ret);
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
            //return SWAP_WORD(((u16*)&boot[location&0x1FFF])[0]);
            r_memcpy16(&ret, &boot[location & 0x1FFF]);
            return(ret);
        }
        if(location >= 0x03f00000 && location <  0x03f0002f)
        {
            //return SWAP_WORD(((u16*)&RDram_Registers[location&0xFF])[0]);
            r_memcpy16(&ret, &RDram_Registers[location & 0xff]);
            return(ret);
        }
    }
    if(location <= 0x003fffff)
    {
        //return SWAP_WORD(((u16*)&RDram[location])[0]);
        r_memcpy16(&ret, &RDram[location]);
        return(ret);
    }

    return 0;
}


inline QWORD getqword (unsigned int location)
{
    QWORD ret = 0;

    #ifdef DEBUGMEM
    printf("** Reading from memory location %08X\n", location);
    #endif

    location &= 0x1FFFFFFF;               // Everything WRAPS!!!

    if(location >= 0x1FC00000 && location <= 0x1FC007FF)
    {
        //return SWAP_DWORD(((u32*)&PIF_RxM[location&0x7FF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&PIF_RxM[location&0x7FF])[0]) << 32);
    r_memcpy64(&ret, &PIF_RxM[location & 0x7FF]);
    return(ret);
    }
    if(location >= 0x10000000)
    {
    //    return SWAP_DWORD(((u32*)&g_n64rom[location&0x0FFFFFFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&g_n64rom[location&0x0FFFFFFF])[0]) << 32);
    r_memcpy64(&ret, &g_n64rom[location & 0x0FFFFFFF]);
    return(ret);
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
            //return SWAP_DWORD(((u32*)&SI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&SI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &SI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
        //    return SWAP_DWORD(((u32*)&RI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&RI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &RI_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04600000 && location <= 0x04600033)
        {
         //   return SWAP_DWORD(((u32*)&PI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&PI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &PI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
           // return SWAP_DWORD(((u32*)&AI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&AI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &AI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
           // return SWAP_DWORD(((u32*)&VI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&VI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &VI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
         //   return SWAP_DWORD(((u32*)&MI_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&MI_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &MI_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04200000 && location <= 0x0420000F)
        {
         //   return SWAP_DWORD(((u32*)&DP_Span_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&DP_Span_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &DP_Span_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
        //    return SWAP_DWORD(((u32*)&DP_Cmd_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&DP_Cmd_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &DP_Cmd_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
        //    return SWAP_DWORD(((u32*)&boot[location&0x1FFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&boot[location&0x1FFF])[0]) << 32);
        r_memcpy64(&ret, &boot[location & 0x1FFF]);
        return(ret);
        }
        if(location >= 0x03f00000 && location <  0x03f0002f)
        {
        //    return SWAP_DWORD(((u32*)&RDram_Registers[location&0xFF])[1]) | ((QWORD)SWAP_DWORD(((u32*)&RDram_Registers[location&0xFF])[0]) << 32);
        r_memcpy64(&ret, &RDram_Registers[location & 0xff]);
        return(ret);
        }
    }
    if(location <= 0x003fffff)
    {
     //   return SWAP_DWORD(((u32*)&RDram[location])[1]) | ((QWORD)SWAP_DWORD(((u32*)&RDram[location])[0]) << 32);
    r_memcpy64(&ret, &RDram[location]);
    return(ret);
    }

    return 0;
}


inline BYTE getbyte (unsigned int location)
{
    BYTE ret = 0;

    #ifdef DEBUGMEM
    printf("** Reading from memory location %08X\n", location);
    #endif

    location &= 0x1FFFFFFF; // Everything WRAPS!!!

    if(location >= 0x1FC00000 && location <= 0x1FC007FF)
    {
    //    return ((u8*)&PIF_RxM[location&0x7FF])[0];
    r_memcpy8(&ret, &PIF_RxM[location & 0x7FF]);
    return(ret);
    }
    if(location >= 0x10000000)
    {
    //    return ((u8*)&g_n64rom[location&0x0FFFFFFF])[0];
    r_memcpy8(&ret, &g_n64rom[location & 0x0FFFFFFF]);
    return(ret);
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
        //    return ((u8*)&SI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &SI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
        //    return ((u8*)&RI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &RI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04600000 && location <= 0x04600033)
        {
        //    return ((u8*)&PI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &PI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
        //    return ((u8*)&AI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &AI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
        //    return ((u8*)&VI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &VI_Registers[location & 0xff]);
        return(ret);
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
        //    return ((u8*)&MI_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &MI_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04200000 && location <= 0x0420000F)
        {
         //   return ((u8*)&DP_Span_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &DP_Span_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
        //    return ((u8*)&DP_Cmd_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &DP_Cmd_Registers[location & 0xfF]);
        return(ret);
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
        //    return ((u8*)&boot[location&0x1FFF])[0];
        r_memcpy8(&ret, &boot[location & 0x1FFF]);
        return(ret);
        }
        if(location >= 0x03f00000 && location <  0x03f0002f)
        {
        //    return ((u8*)&RDram_Registers[location&0xFF])[0];
        r_memcpy8(&ret, &RDram_Registers[location & 0xff]);
        return(ret);
        }
    }
    if(location <= 0x003FFFFF)
    {
     //   return ((u8*)&RDram[location])[0];
    r_memcpy8(&ret, &RDram[location]);
    return(ret);
    }

    return 0;
}

// Write to Memory
//////////////////////////////////////////////////////////////////////////
inline void putbyte (unsigned int location, BYTE data)
{
    location &= 0x1FFFFFFF;               // Everything WRAPS!!!

    if(location >= 0x1FC007C0 && location <= 0x1FC007FF)
    {
        ((u8*)&PIF_RxM[location&0x7FF])[0] = data;
        //r_memcpy8(&PIF_RxM[location & 0x7FF], &data);
        //return;
    }

    if(location >= 0x10000000)
    {
        return; // I don't really think we can do this!
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
            ((u8*)&SI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&SI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
            ((u8*)&RI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&RI_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04600000 && location <= 0x04600034)
        {
            switch (location & 0xFF)
            {
                case 0x00:
                    PI.DRAM = data;
                break;

                case 0x04:
                    PI.Cart = data;
                break;

                case 0x08:
                    PI.RD_LEN = data;
                break;

                case 0x0C:
                    PI.WR_LEN = data;
                    Do_PI_Transfer();
                break;

                case 0x10:
                    if(data < 2)
                    ((u8*)&PI_Registers[location&0xFF])[0] = data;
                    //r_memcpy8(&PI_Registers[location & 0xff], &data);
                return;
            }

            ((u8*)&PI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&PI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
            ((u8*)&AI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&AI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
            ((u8*)&VI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&VI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
            ((u8*)&MI_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&MI_Registers[location & 0xFf], &data);
            return;
        }
        if(location >= 0x04200000 && location <= 0x0420000F)
        {
            ((u8*)&DP_Span_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&DP_Span_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
            ((u8*)&DP_Cmd_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&DP_Cmd_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
            ((u8*)&boot[location&0x1FFF])[0] = data;
            //r_memcpy8(&boot[location & 0x1FFF], &data);
            return;
        }
        if(location >= 0x03F00000 && location <= 0x03F0002F)
        {
            ((u8*)&RDram_Registers[location&0xFF])[0] = data;
            //r_memcpy8(&RDram_Registers[location & 0xff], &data);
            return;
        }
    }
    if(location <= 0x003FFFFF)
        ((u8*)&RDram[location&0x003FFFFF])[0] = data;
    //r_memcpy8(&RDram[location & 0x003fffff], &data);
}

inline void putdword (unsigned int location, DWORD data)
{
    location &= 0x1FFFFFFF;               // Everything WRAPS!!!

    if(location >= 0x1FC007C0 && location <= 0x1FC007FF)
    {
        //((u32*)&PIF_RxM[location&0x7FF])[0] = SWAP_DWORD(data);
        r_memcpy32(&PIF_RxM[location & 0x7FF], &data);
        return;
    }
    if(location >= 0x10000000)
    {
        return; // I don't really think we can do this!
    }
    if(location >= 0x03f00000 && location <= 0x04900000)
    {
        if(location >= 0x04800000 && location <= 0x0480001B)
        {
            //((u32*)&SI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&SI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04700000 && location <= 0x0470001F)
        {
            //((u32*)&RI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&RI_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04600000 && location <= 0x04600034)
        {
            switch (location & 0xFF)
            {
                case 0x00:
                    PI.DRAM = data;
                break;

                case 0x04:
                    PI.Cart = data;
                break;

                case 0x08:
                    PI.RD_LEN = data;
                break;

                case 0x0C:
                    PI.WR_LEN = data;
                    Do_PI_Transfer();
                break;

                case 0x10:
                    if(data < 2)
                        //((u32*)&PI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
                        r_memcpy32(&PI_Registers[location & 0xff], &data);
                return;
            }
            //((u32*)&PI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&PI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04500000 && location <= 0x04500017)
        {
            //((u32*)&AI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&AI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04400000 && location <= 0x04400037)
        {
            //((u32*)&VI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&VI_Registers[location & 0xff], &data);
            return;
        }
        if(location >= 0x04300000 && location <= 0x0430000F)
        {
            //((u32*)&MI_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&MI_Registers[location & 0xFf], &data);
            return;
        }
        if(location >= 0x04200000 && location <= 0x0420000F)
        {
            //((u32*)&DP_Span_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&DP_Span_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04100000 && location <= 0x0410001F)
        {
            //((u32*)&DP_Cmd_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&DP_Cmd_Registers[location & 0xfF], &data);
            return;
        }
        if(location >= 0x04000000 && location <  0x04002000)
        {
            //((u32*)&boot[location&0x1FFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&boot[location & 0x1FFF], &data);
            return;
        }
        if(location >= 0x03F00000 && location <= 0x03F0002F)
        {
            //((u32*)&RDram_Registers[location&0xFF])[0] = SWAP_DWORD(data);
            r_memcpy32(&RDram_Registers[location & 0xff], &data);
            return;
        }
    }

    if(location <= 0x003FFFFF)
        //((u32*)&RDram[location&0x003FFFFF])[0] = SWAP_DWORD(data);
        r_memcpy32(&RDram[location & 0x003fffff], &data);
    return;
}

inline void putqword (unsigned int location, QWORD data)
{
  location &= 0x1FFFFFFF;               // Everything WRAPS!!!


  if(location >= 0x1FC007C0 && location <= 0x1FC007FF)
  {
    r_memcpy64(&PIF_RxM[location & 0x7FF], &data);
    return;
  }
  if(location >= 0x10000000)
  {
    return; // I don't really think we can do this!
  }
  if(location >= 0x03f00000 && location <= 0x04900000)
  {
    if(location >= 0x04800000 && location <= 0x0480001B)
    {
      r_memcpy64(&SI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04700000 && location <= 0x0470001F)
    {
      r_memcpy64(&RI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04600000 && location <= 0x04600034)
    {
      switch (location & 0xff)
      {
        case 0x00:
          PI.DRAM = data;
          break;
        case 0x04:
          PI.Cart = data;
          break;
        case 0x08:
          PI.RD_LEN = data;
          break;
        case 0x0C:
          PI.WR_LEN = data;
          Do_PI_Transfer();
          break;
        case 0x10:
          if(data < 2)
            r_memcpy64(&PI_Registers[location & 0xff], &data);
          return;
      }
      r_memcpy64(&PI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04500000 && location <= 0x04500017)
    {
      r_memcpy64(&AI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04400000 && location <= 0x04400037)
    {
      r_memcpy64(&VI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04300000 && location <= 0x0430000F)
    {
      r_memcpy64(&MI_Registers[location & 0xfF], &data);
      return;
    }
    if(location >= 0x04200000 && location <= 0x0420000F)
    {
      r_memcpy64(&DP_Span_Registers[location & 0xfF], &data);
      return;
    }
    if(location >= 0x04100000 && location <= 0x0410001F)
    {
      r_memcpy64(&DP_Cmd_Registers[location & 0xfF], &data);
      return;
    }
    if(location >= 0x04000000 && location <  0x04002000)
    {
      r_memcpy64(&boot[location & 0x1FFF], &data);
      return;
    }
    if(location >= 0x03f00000 && location <= 0x03f0002f)
    {
      r_memcpy64(&RDram_Registers[location & 0xff], &data);
      return;
    }
  }
  if(location <= 0x003fffff)
    r_memcpy64(&RDram[location & 0x003fffff], &data);
  return;
}

inline void putword (unsigned int location, WORD data)
{
  location &= 0x1FFFFFFF;               // Everything WRAPS!!!


  if(location >= 0x1FC007C0 && location <= 0x1FC007FF)
  {
    r_memcpy16(&PIF_RxM[location & 0x7FF], &data);
    return;
  }
  if(location >= 0x10000000)
  {
    return; // I don't really think we can do this!
  }
  if(location >= 0x03f00000 && location <= 0x04900000)
  {
    if(location >= 0x04800000 && location <= 0x0480001B)
    {
      r_memcpy16(&SI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04700000 && location <= 0x0470001F)
    {
      r_memcpy16(&RI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04600000 && location <= 0x04600034)
    {
      switch (location & 0xff)
      {
        case 0x00:
          PI.DRAM = data;
          break;
        case 0x04:
          PI.Cart = data;
          break;
        case 0x08:
          PI.RD_LEN = data;
          break;
        case 0x0C:
          PI.WR_LEN = data;
          Do_PI_Transfer();
          break;
        case 0x10:
          if(data < 2)
            r_memcpy16(&PI_Registers[location & 0xff], &data);
          return;
      }
      r_memcpy16(&PI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04500000 && location <= 0x04500017)
    {
      r_memcpy16(&AI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04400000 && location <= 0x04400037)
    {
      r_memcpy16(&VI_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04300000 && location <= 0x0430000F)
    {
      r_memcpy16(&MI_Registers[location & 0xfF], &data);
      return;
    }
    if(location >= 0x04200000 && location <= 0x0420000F)
    {
      r_memcpy16(&DP_Span_Registers[location & 0xfF], &data);
      return;
    }
    if(location >= 0x04100000 && location <= 0x0410001F)
    {
      r_memcpy16(&DP_Cmd_Registers[location & 0xff], &data);
      return;
    }
    if(location >= 0x04000000 && location <  0x04002000)
    {
      r_memcpy16(&boot[location & 0x1FFF], &data);
      return;
    }
    if(location >= 0x03f00000 && location <= 0x03f0002f)
    {
      r_memcpy16(&RDram_Registers[location & 0xff], &data);
      return;
    }
  }
  if(location <= 0x003fffff)
    r_memcpy16(&RDram[location & 0x003fffff], &data);
  return;
}
