#include "rom.h"

FILE *romfp;
FILE *g_logfile;
ROMHeader n64header;
struct regs_typ Nregs;

int LoadRom(const char* file)
{
	QWORD current_read;
	unsigned short workvar, workvar2;
    unsigned long int i = 0, filesize = 0;

#ifdef DEBUG
    g_logfile = fopen("memory.dmp", "wb");
#endif
    if((romfp = fopen(file, "rb")) == NULL)
	{
		printf("Can't open the rom.\n");
        return -1;
	}

#ifdef GP2X
    gp2x_memset(&n64header, 0, sizeof(n64header));
#else
	memset(&n64header, 0, sizeof(n64header));
#endif

	fread(&n64header.valid, 2, 1, romfp);
	fread (&workvar, 2, 1, romfp);
	n64header.is_compressed = workvar>>8;
	n64header.unknown = workvar & 0x00FF;

	fread (&n64header.Clockrate, 4, 1, romfp);
	SwapWords(&n64header.Clockrate);

	fread (&n64header.Program_Counter, 4, 1, romfp);
	SwapWords(&n64header.Program_Counter);

	fread(&n64header.Release, 4, 1, romfp);
	SwapWords(&n64header.Release);

	fread(&n64header.CRC1, 4, 1, romfp);
	SwapWords(&n64header.CRC1);

	fread(&n64header.CRC2, 4, 1, romfp);
	SwapWords(&n64header.CRC2);

	fread (&n64header.filler1, 8, 1, romfp);
	fread(n64header.Image_Name, 20, 1, romfp);
	fread(&current_read, 8, 1, romfp);
	fread(n64header.filler2, 7, 1, romfp);
	fread(&n64header.Manu_ID, 1, 1, romfp);
	fread(&n64header.Cart_ID, 2, 1, romfp);
	fread(&n64header.Country_Code, 1, 1, romfp);
	fread(&n64header.filler3, 1, 1, romfp);

    printf ("Validation word = %4X\n", n64header.valid);
	if(n64header.is_compressed == 0x12)
        printf ("The file is not compressed\n");
    else
        if(n64header.is_compressed == 0x13)
            printf ("The file is compressed\n");
    printf ("%X\n", n64header.is_compressed);
    printf ("%X\n", n64header.unknown);
	printf ("The Clockrate is: %x\n", n64header.Clockrate);
    printf ("Starting Program Counter:  %8X\n", n64header.Program_Counter);
	printf ("Release version: %x\n", n64header.Release);
    printf ("CRC1: %8X\n", n64header.CRC1);
    printf ("CRC2: %8X\n", n64header.CRC2);
	printf ("filler (should be all 0's): %x \n", n64header.filler1);
	printf ("Image Name: %s\n", n64header.Image_Name);
	printf ("filler2 (should be all 0's): %x\n", n64header.filler2);
	printf ("Manufacturers ID: %x\n", n64header.Manu_ID);

	printf ("Cartridge ID: %x\n", n64header.Cart_ID);
	printf ("Country Code: %x\n", n64header.Country_Code);
	printf ("filler3 (should be 0): %x\n", n64header.filler3);
	printf ("Boot code starts here (1008 dwords)\n");
	printf ("Game Code starts at file ofs 1000h\n");

    fseek(romfp, 0, SEEK_END);
    filesize = ftell(romfp);

    g_n64rom = (char*)malloc(filesize);
    if((g_n64rom == NULL))
	{
		printf ("Insufficient memory to load ROM image\n");
		return -1;
	}

#ifdef GP2X
    gp2x_memset(&n64header, 0, sizeof(n64header));
#else
    memset(&g_n64rom[0], 0, filesize);
#endif

    fseek(romfp, 0, SEEK_SET);
    for (i = 0; i < filesize/2; i += 2)
    {
        fread(&g_n64rom[i], 2, 1, romfp);
        SwapBytes((WORD *)&g_n64rom[i]);
    }

#ifdef GP2X
    gp2x_memcpy(boot, g_n64rom, 0x1000);
#else
    memcpy(boot, g_n64rom, 0x1000);
#endif
    SP_Registers = &boot[0];

    Nregs.PC = 0;
    Nregs.REAL_PC = n64header.Program_Counter;

#ifdef GP2X
    gp2x_memset(&n64header, 0, sizeof(n64header));
#else
    memset(&Nregs.main.regs, 0, sizeof(Nregs.main.regs));
    memset(&Nregs.cop0.regs, 0, sizeof(Nregs.cop0.regs));
#endif

    return 0;
}

void CloseRom()
{
	free(g_n64rom);
    fclose(romfp);
#ifdef DEBUG
    fclose(g_logfile);
#endif
}
