#ifdef GP2X
    #include "minimal.h"
#else
    #include <SDL/SDL.h>
#endif
#include <time.h>

#include "video.h"

#ifndef GP2X
SDL_Surface *surface = NULL;
SDL_Event event;
#endif

char buffer[20];
u16 *pix;

void (*DoVideo)();
void displayFPS(void);

#ifdef GP2X
void VMLIB(void)
{
    unsigned long  pad = gp2x_joystick_read();
    if(pad&GP2X_START)
        loop = 0;

    // quick optimization
    DWORD tmp1 = getdword(0x04400024);
    DWORD tmp2 = getdword(0x04400028);

    int S16=1;
    int i,j;
    int h_end = tmp1 & 0x3FF;
    int h_start = (tmp1 >> 16) & 0x3FF;
    int v_end = tmp2& 0x3FF;
    int v_start = (tmp2 >> 16) & 0x3FF;
    int scale_x = ((int)getdword(0x04400030) & 0xFFF) >> 16; /// 512;
    int scale_y = (((int)getdword(0x04400034) & 0xFFF)>>1) >> 16; /// 512;

    int minx = (320-(h_end-h_start));
    int maxx = 320-minx;
    int miny = (240-(v_end-v_start));
    int maxy = 240-miny;
    int px, py;
    py=0;

	for (j=0; j<240; j++)
    {
        px=0;
        for (i=0; i<320; i++)
        {
            int iOffset = (j*640) + (i<<1);
            WORD uColor = getword( getdword(0x04400004) + iOffset );
            gp2x_video_RGB[0].screen16[j*320+i] = uColor;
            px += scale_x;
        }
        py += scale_y;
    }

    displayFPS();

    gp2x_video_RGB_flip(0);
}
#else
void VSDL(void)
{
    while(SDL_PollEvent(&event))
    {
        switch(event.type)
        {
            case SDL_QUIT:
                loop = 0;
            break;

            case SDL_KEYDOWN:
                switch(event.key.keysym.sym)
                {
                    case SDLK_ESCAPE:
                        loop = 0;
                    break;
                }
            break;
        }
    }

    // quick optimization
    DWORD tmp1 = getdword(0x04400024);
    DWORD tmp2 = getdword(0x04400028);

    int S16=1;
    int i,j;
    int h_end = tmp1 & 0x3FF;
    int h_start = (tmp1 >> 16) & 0x3FF;
    int v_end = tmp2& 0x3FF;
    int v_start = (tmp2 >> 16) & 0x3FF;
    int scale_x = ((int)getdword(0x04400030) & 0xFFF) >> 16; /// 512;
    int scale_y = (((int)getdword(0x04400034) & 0xFFF)>>1) >> 16; /// 512;

    int minx = (320-(h_end-h_start));
    int maxx = 320-minx;
    int miny = (240-(v_end-v_start));
    int maxy = 240-miny;
    int px, py;
    py=0;

	for (j=0; j<240; j++)
    {
        px=0;
        for (i=0; i<320; i++)
        {
            int iOffset = (j*640) + (i<<1);
            WORD uColor = getword( getdword(0x04400004) + iOffset );
            pix[j*320+i] = uColor;
            px += scale_x;
        }
        py += scale_y;
    }

    displayFPS();

	SDL_Flip(surface);
}
#endif

time_t lastTime;

void displayFPS(void)
{
    static int fps = 0;
    time_t currentTime;
    time(&currentTime);

    ++fps;

    if(difftime(currentTime,lastTime) > 1.0f )
    {
        lastTime = currentTime;
        #ifdef GP2X
        gp2x_printf(NULL,0,0, "FPS: %i : Rapide non ? =)", (int)fps);
        #else
        sprintf(buffer,"FPS: %i",(int)fps);
        SDL_WM_SetCaption(buffer,NULL);
        #endif
        fps = 0;
    }
}

void InitVideo(void)
{
    time(&lastTime);
    #ifndef GP2X
    DoVideo = VSDL;
    #else
    DoVideo = VMLIB;
    #endif

#ifdef GP2X
    gp2x_init(1000, 16, 22050, 16, 1, 60, 0);
    gp2x_video_logo_enable(1);
#else
    SDL_Init(SDL_INIT_VIDEO);
	surface = SDL_SetVideoMode(320, 240, 16, SDL_SWSURFACE/* | SDL_DOUBLEBUF*/);

	SDL_ShowCursor(0);

	pix = (u16 *)surface->pixels;
#endif
}

#ifdef GP2X
void gp2x_sound_frame(void *blah, void *buff, int samples){}
#endif
